@@ -1,7 +1,7 @@
 package App::instopt;
 
-our $DATE = '2019-10-09'; # DATE
-our $VERSION = '0.012'; # VERSION
+our $DATE = '2019-10-15'; # DATE
+our $VERSION = '0.013'; # VERSION
 
 use 5.010001;
 use strict 'subs', 'vars';
@@ -13,10 +13,13 @@ use File::chdir;
 use File::MoreUtil qw(dir_has_non_dot_files);
 use Perinci::Object;
 use PerlX::Maybe;
+use Sah::Schema::software::arch;
 
 use vars '%Config';
 our %SPEC;
 
+our @all_known_archs = @{ $Sah::Schema::software::arch::schema->[1]{in} };
+
 our %args_common = (
     download_dir => {
         schema => 'dirname*',
@@ -94,9 +97,12 @@ _
 );
 
 sub _set_args_default {
-    my $args = shift;
-    if (!$args->{arch}) {
-        $args->{arch} = App::swcat::_detect_arch();
+    my ($args, $opts) = @_;
+
+    if ($opts->{set_default_arch}) {
+        if (!$args->{arch}) {
+            $args->{arch} = App::swcat::_detect_arch();
+        }
     }
     if (!$args->{download_dir}) {
         require PERLANCAR::File::HomeDir;
@@ -177,10 +183,13 @@ sub _convert_download_urls_to_filenames {
 }
 
 sub _init {
-    my ($args) = @_;
+    my ($args, $opts) = @_;
+
+    $opts //= {};
+    $opts->{set_default_arch} //= 1;
 
     unless ($App::instopt::state) {
-        _set_args_default($args);
+        _set_args_default($args, $opts);
         my $state = {
         };
         $App::instopt::state = $state;
@@ -323,15 +332,20 @@ $SPEC{list_downloaded} = {
     summary => 'List all downloaded software',
     args => {
         %args_common,
-        detail => {
-            schema => ['bool*', is=>1],
-            cmdline_aliases => {l=>{}},
+        %argopt_arch,
+        %argopt_detail,
+        per_arch => {
+            summary => 'Return per-arch hash in the all_versions field',
+            schema => 'true*',
         },
     },
+    args_rels => {
+        #choose_one => ['arch', 'per_arch'],
+    },
 };
 sub list_downloaded {
     my %args = @_;
-    my $state = _init(\%args);
+    my $state = _init(\%args, {set_default_arch=>0});
 
     my $res = App::swcat::list();
     return [500, "Can't list known software: $res->[0] - $res->[1]"] if $res->[0] != 200;
@@ -353,31 +367,48 @@ sub list_downloaded {
         for my $sw (@$swlist) {
             my $dir = sprintf "%s/%s", substr($sw, 0, 1), $sw;
             unless (-d $dir) {
-                log_trace "Skipping software '$sw': directory doesn't exist";
+                log_trace "Skipping software '$sw': directory '$CWD/$dir' doesn't exist";
                 next SW;
             }
             local $CWD = $dir;
             my $mod = App::swcat::_load_swcat_mod($sw);
-            my @vers;
+            my %arch_vers; # key = arch, val = [ver1, ...]
+            my @archs = defined($args{arch}) ? ($args{arch}) : @all_known_archs;
           VER:
             for my $e (glob "*") {
-                if ($args{_arch}) {
-                    next unless dir_has_non_dot_files("$e/$args{_arch}");
-                } else {
-                    next unless -d $e;
+                unless ($mod->is_valid_version($e)) {
+                    log_trace "Skipping invalid version '$e' of software '$sw'";
+                    next;
+                }
+                for my $arch (@archs) {
+                    #log_trace "Searching software '$sw' version '$e' arch '$arch'";
+                    unless (dir_has_non_dot_files("$e/$arch")) {
+                        #log_trace "Skipping software '$sw' version '$e' arch '$arch': no files found";
+                        next;
+                    }
+                    push @{ $arch_vers{$arch} }, $e;
                 }
-                next unless $mod->is_valid_version($e);
-                push @vers, $e;
             }
-            unless (@vers) {
-                log_trace "Skipping software '$sw': no downloaded versions found";
+
+            my @vers;
+            for my $arch (@archs) {
+                next unless $arch_vers{$arch};
+                log_trace "Found downloaded versions %s for software '%s' arch '$arch'",
+                    $arch_vers{$arch}, $sw, $arch;
+                for my $ver (@{ $arch_vers{$arch} }) {
+                    push @vers, $ver unless grep { $ver eq $_ } @vers;
+                }
             }
             @vers = sort { $mod->cmp_version($a, $b) } @vers;
-            log_trace "Found downloaded versions %s for software '%s'", \@vers, $sw;
+            unless (@vers) {
+                log_trace "Skipping software '$sw': no downloaded versions found";
+                next;
+            }
             push @rows, {
                 software => $sw,
                 latest_version => $vers[-1],
-                all_versions => join(", ", @vers),
+                all_versions => $args{per_arch} ? \%arch_vers : join(", ", @vers),
+                (arch => $args{arch}) x !!defined($args{arch}),
             };
         }
     }
@@ -408,7 +439,7 @@ sub list_downloaded_versions {
 
     return [400, "Please specify software"] unless $args{software};
 
-    my $res = list_downloaded(%args, _software=>$args{software}, _arch=>$args{arch}, detail=>1);
+    my $res = list_downloaded(%args, _software=>$args{software}, arch=>$args{arch}, detail=>1);
     return $res unless $res->[0] == 200;
     my $row = $res->[2][0];
     return [200, "OK (none downloaded)"] unless $row;
@@ -646,32 +677,34 @@ sub cleanup_download_dir {
     require File::Path;
 
     my %args = @_;
-    my $state = _init(\%args);
+    my $state = _init(\%args, {set_default_arch=>0});
 
     local $CWD = $args{download_dir};
-    my $res = list_downloaded(%args, detail=>1);
+    my $res = list_downloaded(%args, detail=>1, per_arch=>1);
     return $res unless $res->[0] == 200;
   SW:
     for my $row (@{ $res->[2] }) {
         my $sw = $row->{software};
         next unless $row->{all_versions};
-        my @vers = split /, /, $row->{all_versions};
-        unless (@vers > 1) {
-            log_trace "Skipping software $sw (<2 versions)";
-            next SW;
-        }
-        pop @vers; # remove latest version
-        my $dir = sprintf "%s/%s", substr($sw, 0, 1), $sw;
-        local $CWD = $dir;
-      VER:
-        for my $v (@vers) {
-            if ($args{-dry_run}) {
-                log_trace "[DRY-RUN] Cleaning up $sw-$v ...";
-            } else {
-                log_trace "Cleaning up software $sw-$v ...";
-                File::Path::remove_tree($v);
+        for my $arch (sort keys %{ $row->{all_versions} }) {
+            my @vers = @{ $row->{all_versions}{$arch} };
+            unless (@vers > 1) {
+                log_trace "Skipping software '$sw' arch '$arch' (<2 versions)";
+                next SW;
             }
-        }
+            pop @vers; # remove latest version
+            my $dir = sprintf "%s/%s", substr($sw, 0, 1), $sw;
+            local $CWD = $dir;
+          VER:
+            for my $v (@vers) {
+                if ($args{-dry_run}) {
+                    log_trace "[DRY-RUN] Cleaning up $sw-$v arch $arch ...";
+                } else {
+                    log_trace "Cleaning up software $sw-$v arch $arch ...";
+                    File::Path::remove_tree("$v/$arch");
+                }
+            }
+        } # for arch
     }
     $args{-dry_run} ? [304, "Dry-run"] : [200];
 }
@@ -896,7 +929,7 @@ App::instopt - Download and install software
 
 =head1 VERSION
 
-This document describes version 0.012 of App::instopt (from Perl distribution App-instopt), released on 2019-10-09.
+This document describes version 0.013 of App::instopt (from Perl distribution App-instopt), released on 2019-10-15.
 
 =head1 SYNOPSIS
 
@@ -1126,12 +1159,18 @@ Arguments ('*' denotes required arguments):
 
 =over 4
 
-=item * B<detail> => I<bool>
+=item * B<arch> => I<software::arch>
+
+=item * B<detail> => I<true>
 
 =item * B<download_dir> => I<dirname>
 
 =item * B<install_dir> => I<dirname>
 
+=item * B<per_arch> => I<true>
+
+Return per-arch hash in the all_versions field.
+
 =item * B<program_dir> => I<dirname>
 
 =back
